function [offset weights] = train_logistic_regression(features, labels, lrate, ...
                                            target_delta, regularization_constant)
    % Gradient ascent algorithm to train logistic regression params

    % offset and weights are the lr params
    offset = 0;
    old_weights = zeros(1, size(features, 2));
    weights = zeros(1, size(features, 2));
    weight_delta = 100000;

    regularizer = ones(1, size(features, 2)) * regularization_constant;
    
    previous_ll = -100000000;
    
    % Using the current weights, calculate the probability that
    % each instance is classified as 1
    weight_sums = repmat(offset, size(features, 1), 1) + ...
        sum(repmat(weights, size(features, 1), 1) .* features, 2);
    probs = exp(weight_sums) ./ (ones(size(features, 1), 1) + ...
                                 exp(weight_sums));
    
    current_ll = sum((labels .* weight_sums) - log(1 + exp(weight_sums))) ...
        - sum(weights .* weights .* regularizer / 2);
    
    fprintf('Training logistic regression. initial ll: :%f \n', current_ll);
    fprintf('(logging messages printed every 100 iterations)\n');
    
    iter_num = 0;
    while(weight_delta > target_delta)
        old_weights = weights;
        gradient = sum(features .* repmat(labels - probs, 1, size(features, 2))) - regularizer .* weights;

        % magnitude limit gradient
        gradient = lrate * gradient;
        iter_num = iter_num + 1;

        % Update weights using the previously calculated label probabilities
        offset = offset + lrate * sum(labels - probs);
        weights = weights + gradient;

        % Using the current weights, calculate the probability that
        % each instance is classified as 1
        weight_sums = repmat(offset, size(features, 1), 1) + ...
            sum(repmat(weights, size(features, 1), 1) .* features, 2);
        probs = exp(weight_sums) ./ (ones(size(features, 1), 1) + ...
                                     exp(weight_sums));

        % Update loglikelihoods.
        previous_ll = current_ll;
        current_ll = sum((labels .* weight_sums) - log(1 + exp(weight_sums))) ...
            - sum(weights .* weights .* regularizer / 2);
        weight_delta = sqrt(sum((old_weights - weights) .* (old_weights - weights)));
        if (mod(iter_num, 100) == 0)
            fprintf('loglikelihood: %f weight delta: %f\n', current_ll, weight_delta);
        end
    end
    
    fprintf('final ll: :%f \n', current_ll);